from pydantic.dataclasses import dataclass

from ..base import BaseSamplerConfig


@dataclass
class NormalSamplerConfig(BaseSamplerConfig):
    """NormalSampler config class.

    N/A
    """

    pass
