"""Implementation of a Manifold sampler proposed in (https://arxiv.org/abs/2105.00026).

Available models:
------------------

.. autosummary::
    ~XGen.models.RHVAE
    :nosignatures:

"""

from .rhvae_sampler import RHVAESampler
from .rhvae_sampler_config import RHVAESamplerConfig

__all__ = ["RHVAESampler", "RHVAESamplerConfig"]
