class Pipeline:
    def __init__(self):
        pass

    def __call__(self):
        """
        The method run the Pipeline process and must be implemented in a child class
        """
        raise NotImplementedError()
